(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: nn_manager.ml 239 2010-06-23 16:49:03Z gerd $ *)

(** Namenode component *)

module Bool_var =
  Netplex_cenv.Make_var_type(struct type t = bool end)


let monitor_factory db_conf nn_conf =
  Rpc_netplex.rpc_factory
    ~configure:(fun cf _ -> ())
    ~hooks:(fun conf ->
              ( object
                  inherit Netplex_kit.empty_processor_hooks()
		    
                  method post_add_hook socksrv ctrl =
		    Nn_monitor.post_add_hook socksrv ctrl

		  method post_start_hook cont =
		    let esys = (Netplex_cenv.self_cont()) # event_system in
		    Nn_db.init db_conf esys;
		    Nn_monitor.post_start_hook cont nn_conf

		  method receive_message cont msg args =
		    Nn_monitor.receive_message cont msg args
		end
	      )
	   )
    ~name:"nn_monitor"
    ~setup:(fun _ _ -> ()  (* TODO *) )
    ()


let add_monitor_service (cf, cfaddr, db_conf, nn_conf) socksrv ctrl =
    (* FIXME: There should be better functions for creating these objects in
     Netplex
   *)
  let sockpath =
    ctrl#controller_config#socket_directory ^ "/nn_monitor.rpc" in
  let sockaddr =
    `Socket (Unix.ADDR_UNIX sockpath) in
  let protocol : Netplex_types.protocol =
    ( object
        method name = "RPC"
        method addresses = [| sockaddr |]
        method lstn_backlog = 5
        method lstn_reuseaddr = true
        method so_keepalive = false
        method configure_slave_socket _ = ()
      end
    ) in
  let socksrv_conf : Netplex_types.socket_service_config =
    ( object
        method name = "Nn_monitor"
        method protocols = [protocol]
        method change_user_to = None
        method startup_timeout = (-1.0)
        method controller_config = ctrl#controller_config
      end
    ) in
  let facobj = monitor_factory db_conf nn_conf in
  let processor = 
    facobj # create_processor 
      ctrl#controller_config 
      cf cfaddr in
  let socksrv = 
    Netplex_sockserv.create_socket_service
      processor
      socksrv_conf in
  let wm = 
    Netplex_workload.create_constant_workload_manager
      1 in
  ctrl # add_service socksrv wm


let index_of_this_node nn_conf port =
  let hostname = nn_conf#nn_hostname in
  let nodes = Array.of_list nn_conf#nn_nodes in
  let rec search k =
    if k >= Array.length nodes then
      failwith "Coordinator not found in node list";
    let (h,p) = Plasma_util.parse_host_port nodes.(k) in
    if h = hostname && p = port then
      k
    else
      search(k+1) in
  search 0


let extract_port cont =
  let sock_serv = cont # socket_service # socket_service_config in
  let proto =
    try List.find (fun p -> p#name = "RPC") sock_serv#protocols
    with Not_found ->
      failwith "Cannot find protocol 'RPC'" in
  let addrs = Array.to_list proto#addresses in
  try
    let addr = 
      List.find
	(function
	   | `Socket addr ->
	       ( match addr with
		 | Unix.ADDR_INET(_,p) -> true
		 | _ -> false
	       )
	   | _ ->
	       false
	)
	addrs in
    match addr with
      | `Socket(Unix.ADDR_INET(_,p)) -> p
      | _ -> assert false
  with
    | Not_found ->
	failwith "Cannot determine my own canonical port"


let factory() =
  Rpc_netplex.rpc_factory
    ~configure:(fun cf cfaddr -> 
		  let db_conf = Pfs_db.extract_db_config cf in
		  let nn_conf = Nn_config.extract_node_config cf in
		  (cf, cfaddr, db_conf, nn_conf)
	       )
    ~hooks:(fun ((cf, cfaddr, db_conf, nn_conf) as conf) ->
	      ( object
		  inherit Netplex_kit.empty_processor_hooks()

		  method post_add_hook socksrv ctrl =
		    Nn_datastore_news.post_add_hook socksrv ctrl;
		    add_monitor_service conf socksrv ctrl;
		      (** Note that the monitor is only started on the
                          coordinator node. We don't know this node yet.
		       *)

		  method post_start_hook cont =
		    (** Note that there is only one process, and hence this
                        method runs only once per startup
		     *)
		    let esys = (Netplex_cenv.self_cont()) # event_system in
		    let port = extract_port cont in
		    Nn_db.init db_conf esys;
		    Nn_db.commit_all_prepared_transactions();
		    Nn_datanode_ctrl.post_start_hook ();
		    Netlog.logf `Debug
		      "Waiting for datastore news";
		    ignore(Nn_datastore_news.list ~wait:true ());
		       (* TODO: don't wait here *)
		    Netlog.logf `Debug
		      "Datastore news feed is ready";
		    Nn_notify.post_start_hook();
		    (** Start the election procedure: *)
		    let ee = Nn_elect.elect_e nn_conf port esys in
		    Uq_engines.when_state
		      ~is_done:(fun (is_coord,r) ->
				  (** Reset the datanodes *)
				  if is_coord then (
				    let i = index_of_this_node nn_conf port in
				    Nn_datanode_ctrl.reset();
				    Nn_coord.post_start_hook cont nn_conf i r;
				  )
			       )
		      ee

		  method receive_message cont msg args =
		    Nn_datastore_news.receive_message cont msg args

		end
	      )
	   )
    ~setup:(fun srv (cf, cfaddr, db_conf, nn_conf) ->
	      Nn_elect.setup_elect_api srv nn_conf;
	      match Nn_elect.get_role() with
		| `Coordinator ->
		    Nn_coord.setup srv nn_conf
		| `Slave ->
		    Nn_slave.setup srv nn_conf
		| _ ->
		    ()
	   )
    ~name:"nn_manager"
    ()
